#ifndef DUNE_TECTONIC_LINEARIZATION_HH
#define DUNE_TECTONIC_LINEARIZATION_HH

#include <cstddef>

#include <dune/common/typetraits.hh>
#include <dune/common/hybridutilities.hh>

#include <dune/matrix-vector/algorithm.hh>

#include <dune/istl/matrixindexset.hh>

//#include <dune/tnnmg/functionals/bcqfconstrainedlinearization.hh>

#include "../../utils/debugutils.hh"

/**
 * derived from Dune::TNNMG::BoxConstrainedQuadraticFunctionalConstrainedLinearization<F,BV>
 */
template<class F, class BV>
class Linearization {
public:
  using Matrix = typename F::Matrix;
  using Vector = typename F::Vector;
  using BitVector = BV;
  using ConstrainedMatrix = Matrix;
  using ConstrainedVector = Vector;
  using ConstrainedBitVector = BitVector;

private:
  using This = Linearization<F,BV>;

  template <class T>
  void determineRegularityTruncation(const Vector& x, BitVector& truncationFlags, const T& truncationTolerance)
  {
    for (size_t i = 0; i < x.size(); ++i) {
      if (f_.phi().regularity(i, x[i]) > truncationTolerance) {
        truncationFlags[i] = true;
      }
    }
  }

  template<class NV, class NBV, class T>
  static void determineTruncation(const NV& x, const NV& lower, const NV& upper, NBV&& truncationFlags, const T& truncationTolerance)
  {
    namespace H = Dune::Hybrid;
    H::ifElse(Dune::IsNumber<NV>(), [&](auto id){
      if ((id(x) <= id(lower)+truncationTolerance) || (id(x) >= id(upper) - truncationTolerance))
        id(truncationFlags) = true;
    }, [&](auto id){
      H::forEach(H::integralRange(H::size(id(x))), [&](auto&& i) {
        This::determineTruncation(x[i], lower[i], upper[i], truncationFlags[i], truncationTolerance);
      });
    });
  }

  template<class NV, class NBV>
  static void truncateVector(NV& x, const NBV& truncationFlags)
  {
    namespace H = Dune::Hybrid;
    H::ifElse(Dune::IsNumber<NV>(), [&](auto id){
      if (id(truncationFlags))
        id(x) = 0;
    }, [&](auto id){
      H::forEach(H::integralRange(H::size(id(x))), [&](auto&& i) {
        This::truncateVector(x[i], truncationFlags[i]);
      });
    });
  }

  template<class NM, class RBV, class CBV>
  static void truncateMatrix(NM& A, const RBV& rowTruncationFlags, const CBV& colTruncationFlags)
  {
    namespace H = Dune::Hybrid;
    using namespace Dune::MatrixVector;
    H::ifElse(Dune::IsNumber<NM>(), [&](auto id){
      if(id(rowTruncationFlags) or id(colTruncationFlags))
        A = 0;
    }, [&](auto id){
      H::forEach(H::integralRange(H::size(id(rowTruncationFlags))), [&](auto&& i) {
        auto&& Ai = A[i];
        sparseRangeFor(Ai, [&](auto&& Aij, auto&& j) {
            This::truncateMatrix(Aij, rowTruncationFlags[i], colTruncationFlags[j]);
        });
      });
    });
  }

public:
  Linearization(const F& f, const BitVector& ignore) :
      f_(f),
      ignore_(ignore),
      truncationTolerance_(1e-10)
  {}

  void bind(const Vector& x) {
      //std::cout << "Linearization::bind()" << std::endl;

      auto&& A = f_.quadraticPart();
      auto&& phi = f_.phi();

      // determine which components to truncate
      // ----------------
      truncationFlags_ = ignore_;

      determineTruncation(x, f_.lowerObstacle(), f_.upperObstacle(), truncationFlags_, truncationTolerance_); // obstacle truncation
      determineRegularityTruncation(x, truncationFlags_, truncationTolerance_); // truncation due to regularity deficit of nonlinearity


      // compute hessian
      // ---------------

      // construct sparsity pattern for linearization
      Dune::MatrixIndexSet indices(A.N(), A.M());
      indices.import(A);
      phi.addHessianIndices(indices);

      // construct matrix from pattern and initialize it
      indices.exportIdx(hessian_);
      hessian_ = 0.0;

      // quadratic part
      for (size_t i = 0; i < A.N(); ++i) {
        auto const end = std::end(A[i]);
        for (auto it = std::begin(A[i]); it != end; ++it)
          hessian_[i][it.index()] += *it;
      }

      //print(hessian_, "Hessian:");
      //std::cout << "--------" << (double) hessian_[21][21] << "------------" << std::endl;

      // nonlinearity part
      phi.addHessian(x, hessian_);

      // compute gradient
      // ----------------

      // quadratic part
      negativeGradient_ = derivative(f_)(x); // A*x - b

      // nonlinearity part
      phi.addGradient(x, negativeGradient_);

      // -grad is needed for Newton step
      negativeGradient_ *= -1;

      // truncate gradient and hessian
      truncateVector(negativeGradient_, truncationFlags_);
      truncateMatrix(hessian_, truncationFlags_, truncationFlags_);
  }


  void extendCorrection(ConstrainedVector& cv, Vector& v) const {
    v = cv;
    truncateVector(v, truncationFlags_);
  }

  const BitVector& truncated() const {
    return truncationFlags_;
  }

  const auto& negativeGradient() const {
    return negativeGradient_;
  }

  const auto& hessian() const {
    return hessian_;
  }

private:
  const F& f_;
  const BitVector& ignore_;

  double truncationTolerance_;

  Vector negativeGradient_;
  Matrix hessian_;
  BitVector truncationFlags_;
};

#endif
